import sys
import os
from loguru import logger
from abc import abstractmethod
import re
from typing import List, Dict
import json

here = os.path.dirname(__file__)

sys.path.append(os.path.join(here, ".."))

from ..pipeline import Pipeline  # noqa: E402
from ...datalake import Datalake  # noqa: E402
from ...data_access_service.data_loaders import (  # noqa: E402
    create_data_loader,
)


class RawToBronzePipeline(Pipeline):
    INPUT_DATA_LOADER = "s3"
    OUTPUT_DATA_LOADER = "s3"
    INPUT_LAYER = "ingest"
    OUTPUT_LAYER = "bronze"

    def __init__(self, partition_id: str) -> None:
        self.datalake = Datalake()
        self.partition_id = partition_id
        self.input_loader = create_data_loader(self.INPUT_DATA_LOADER)
        self.output_loader = create_data_loader(self.OUTPUT_DATA_LOADER)
        self.account_ids = self.__get_account_ids__()

    def read_sources(self) -> List[Dict[str, str]]:
        """
        Reads the data from the data sources.

        :return: The data from the data sources.
        """
        logger.info("Reading data from datalake.")
        return [
            self.input_loader.read(
                self.datalake[self.INPUT_LAYER][self.INPUT_KEY],
                {
                    "partition_id": self.partition_id,
                    "account_id": account_id,
                },
            )
            for account_id in self.account_ids
        ]

    def write_sources(self, transformed_data: List[Dict[str, str]]):
        """
        Writes the data to the data sources.

        :param transformed_df: The transformed data.
        """
        logger.info("Writing data to datalake.")
        """
        Writes the data to the data sources.
        """
        datalake_config = self.datalake[self.OUTPUT_LAYER][self.OUTPUT_KEY]
        for account_id in self.account_ids:
            partition_config = {
                "partition_id": self.partition_id,
                "account_id": account_id,
            }
            data = transformed_data[account_id]
            self.output_loader.write(
                json.dumps(data), datalake_config, partition_config
            )

    @abstractmethod
    def transform(
        self, source_data: List[Dict[str, str]]
    ) -> List[Dict[str, str]]:
        """
        Transforms the data.

        :param source_df: The data to transform.
        :return: The transformed data.
        """
        raise NotImplementedError("Transform method not implemented!")

    def run(self):
        """
        Runs the pipeline.
        """
        logger.info("Running the bronze to silver pipeline...")
        source_data = self.read_sources()
        transformed_data = self.transform(source_data)
        self.write_sources(transformed_data)

    def __get_account_ids__(self) -> List[str]:
        """
        Gets the account ids from s3 bucket for the given partition
        by getting the paths of the files in the bucket.

        :return: The account ids.
        """
        s3_client = self.input_loader.s3_connection.s3_client
        contents = self.input_loader.get_all_bucket_objects(
            s3_client,
            self.datalake[self.INPUT_LAYER][self.INPUT_KEY],
            {"partition_id": self.partition_id},
        )
        return [
            re.search(r"account_id=(.*?)/", content["Key"]).group(1)
            for content in contents
        ]

    @staticmethod
    def camel_to_snake(s: str) -> str:
        name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()



from .data_loader import DataLoader
from ..data_connections import connect
from loguru import logger
from typing import Dict
import json


class S3DataLoader(DataLoader):
    NAME = "s3"

    def __init__(self):
        """
        Constructor for S3DataLoader class.
        """
        self.s3_connection = connect(self.NAME)

    def _build_s3_prefix(
        self, datalake_config: Dict[str, str], partition_config: Dict[str, str]
    ):
        """
        Method for building the S3 prefix.

        :param datalake_config: The configuration of the datalake to read from.
        :param partition_id: The partition of the datalake to read from.
        """
        return "{0}/{1}".format(
            datalake_config["datalake_key"],
            self.build_partition_path(partition_config),
        )

    def get_all_bucket_objects(
        self,
        s3_client,
        datalake_config: Dict[str, str],
        partition_config: Dict[str, str],
    ):
        """
        Method for getting all the objects in the S3 bucket.

        :param s3_client: The S3 client.
        :param datalake_config: The configuration of the datalake to read from.
        """
        prefix = self._build_s3_prefix(datalake_config, partition_config)
        response = s3_client.list_objects_v2(
            Bucket=datalake_config["datalake_bucket"],
            Prefix=prefix,
        )
        return response["Contents"]

    def _build_s3_file_key(self, key: str, file_extension: str):
        """
        Method for building the S3 file key.

        :param key: The key of the datalake to read from.
        :param file_extension: The file extension of the file to read.
        """
        return "{0}.{1}".format(key, file_extension)

    def read(
        self, datalake_config: Dict[str, str], partition_config: Dict[str, str]
    ):
        """
        Method for reading data from the S3 bucket.

        :param datalake_layer: The layer of the datalake to read from.
        :param datalake_key: The key of the datalake to read from.
        """
        logger.info("Reading data from S3...")
        s3_client = self.s3_connection.s3_client
        contents = self.get_all_bucket_objects(
            s3_client, datalake_config, partition_config
        )
        # read the content of each file
        output = {}
        for file in contents:
            key = file["Key"]
            response = s3_client.get_object(
                Bucket=datalake_config["datalake_bucket"],
                Key=key,
            )
            json_data = response["Body"].read().decode("utf-8")
            parsed_data = json.loads(json_data)
            output[key] = parsed_data
        logger.info("Finished reading data from S3!")
        return output

    def write(
        self,
        file_content,
        datalake_config: Dict[str, str],
        partition_config: Dict[str, str],
    ):
        """
        Method for writing data to the S3 bucket.

        :param datalake_layer: The layer of the datalake to write to.
        :param datalake_key: The key of the datalake to write to.
        """
        logger.info("Writing data to S3...")
        s3_client = self.s3_connection.s3_client
        prefix = self._build_s3_prefix(datalake_config, partition_config)
        key = self._build_s3_file_key(
            datalake_config["datalake_key"], datalake_config["file_extension"]
        )
        s3_client.put_object(
            Bucket=datalake_config["datalake_bucket"],
            Key="{0}/{1}".format(prefix, key),
            Body=file_content,
        )



import sys
import os
from loguru import logger
from pyspark.sql import DataFrame as SparkDataFrame
import pyspark.sql.functions as F
from typing import Tuple

here = os.path.dirname(__file__)

sys.path.append(os.path.join(here, ".."))

from .bronze_to_silver_pipeline import BronzeToSilverPipeline  # noqa: E402


class BronzeToSilverBalancesPipeline(BronzeToSilverPipeline):
    INPUT_DATA_LOADER = "spark_s3"
    OUTPUT_DATA_LOADER = "spark_s3"
    INPUT_LAYER = "bronze"
    OUTPUT_LAYER = "silver"
    INPUT_KEY = "balances"
    OUTPUT_KEY = "balances"

    def read_source_exchange_rates(self) -> SparkDataFrame:
        """
        Reads the exchange rates from the data sources.

        :return: The exchange rates from the data sources.
        """
        logger.info("Reading exchange rates from datalake.")
        exchange_rates_df = self.input_loader.read(
            self.datalake["master"]["exchange_rates"],
            {"partition_id": self.partition_id},
        )
        return exchange_rates_df

    def read_sources(self) -> Tuple[SparkDataFrame, SparkDataFrame]:
        """
        Reads the data from the data sources.

        :return: The data from the data sources.
        """
        logger.info("Reading data from datalake.")
        source_df = self.input_loader.read(
            self.datalake[self.INPUT_LAYER][self.INPUT_KEY],
            {
                "partition_id": self.partition_id,
            },
        )
        currency_rates_df = self.read_source_exchange_rates()
        return source_df, currency_rates_df

    def transform(
        self, source_df: SparkDataFrame, currency_rates_df: SparkDataFrame
    ) -> SparkDataFrame:
        """
        Transforms the data.

        :param source_df: The data to transform.
        :return: The transformed data.
        """
        logger.info("Transforming data.")
        transformed_df = self.convert_currencies(source_df, currency_rates_df)
        transformed_df = transformed_df.toDF(
            *[
                "amount",
                "currency",
                "type",
                "partition_id",
                "account_id",
                "amount_PLN",
            ]
        )
        return transformed_df

    def convert_currencies(
        self,
        source_df: SparkDataFrame,
        currency_rates: SparkDataFrame,
        base_currency: str = "PLN",
    ) -> SparkDataFrame:
        """
        Converts currencies.

        :param source_df: The data to transform.
        :param currency_rates: The currency rates.
        :return: The transformed data.
        """
        logger.info("Converting currencies.")
        transformed_df = (
            source_df.join(
                currency_rates,
                source_df.balance_currency == currency_rates.currency,
                "left",
            )
            .withColumn(
                f"amount_{base_currency}",
                F.round(source_df.balance_amount / currency_rates.rate, 2),
            )
            .drop("rate", "currency")
        )
        return transformed_df

    def run(self):
        """
        Runs the pipeline.
        """
        logger.info("Running the bronze to silver pipeline...")
        source_df, currency_rates_df = self.read_sources()
        transformed_df = self.transform(source_df, currency_rates_df)
        self.write_sources(transformed_df)



from loguru import logger
import boto3
import json
import re
from typing import Dict, Tuple
import urllib


def lambda_handler(event, context):
    (
        source_bucket,
        source_key,
        partition_id,
        account_id,
        data_type,
    ) = get_ingestion_details(event)
    formatter = Formatter(data_type)
    destination_bucket = "budget-guard-bronze"
    destination_key = f"{data_type}/partition_id={partition_id}/account_id={account_id}/{source_key.split('/')[-1]}"  # noqa
    data = s3_read_json(source_bucket, source_key)
    data = formatter(data)
    s3_write_json(data, destination_bucket, destination_key)
    logger.info(
        "Successfully copied data from {0} to {1}!".format(
            source_bucket, destination_bucket
        )
    )


def get_ingestion_details(event: Dict) -> Tuple[str, str, str, str, str]:
    """
    Method for getting the details of the ingestion event.
    """

    source_bucket = event["Records"][0]["s3"]["bucket"]["name"]
    source_key = urllib.parse.unquote(
        event["Records"][0]["s3"]["object"]["key"]
    )
    logger.info(
        "Received event from {0} for {1}.".format(source_bucket, source_key)
    )  # noqa
    partition_id = re.search(r"partition_id=(\d+)", source_key).group(1)
    account_id = re.search(r"account_id=([a-zA-Z0-9-]+)", source_key).group(1)
    data_type = source_key.split("/")[0]
    logger.info(
        "Received event from {0} for {1} data in partition {2} for account {3}.".format(  # noqa
            source_bucket, data_type, partition_id, account_id
        )
    )

    return source_bucket, source_key, partition_id, account_id, data_type


def s3_read_json(source_bucket: str, source_key: str) -> Dict:
    """
    Method for reading a JSON file from S3.
    """

    # Create an S3 client
    s3 = boto3.client("s3")

    try:
        logger.info("Reading JSON data from {0}...".format(source_bucket))
        response = s3.get_object(Bucket=source_bucket, Key=source_key)
        json_data = response["Body"].read().decode("utf-8")
        parsed_data = json.loads(json_data)
        return parsed_data

    except Exception as e:
        print("Error: {}".format(str(e)))
        return {"statusCode": 500, "body": "Error: {}".format(str(e))}


def s3_write_json(json_data: str, target_bucket: str, target_key: str) -> Dict:
    """
    Method for writing a JSON file to S3.
    """

    # Create an S3 client
    s3 = boto3.client("s3")

    try:
        logger.info("Writing JSON data to {0}...".format(target_bucket))
        response = s3.put_object(
            Bucket=target_bucket, Key=target_key, Body=json.dumps(json_data)
        )

        return response

    except Exception as e:
        print("Error: {}".format(str(e)))
        return {"statusCode": 500, "body": "Error: {}".format(str(e))}


class Formatter:
    def __init__(self, data_type: str):
        self.data_type = data_type

    def __call__(self, data: Dict) -> Dict:
        return self.get_nordigen_function(self.data_type)(data)

    @staticmethod
    def camel_to_snake(s: str) -> str:
        name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()

    def get_nordigen_function(self, data_type: str):
        if data_type == "balances":
            return self.format_balances
        elif data_type == "transactions":
            return self.format_transactions
        elif data_type == "details":
            return self.format_details
        elif data_type == "metadata":
            return self.format_metadata
        else:
            raise ValueError("Invalid data type specified!")

    def format_transactions(self, transactions):
        transactions_flatten = []
        logger.info("Flattening transactions...")
        for transaction_status, transactions_info in transactions[
            "transactions"
        ].items():
            logger.info(
                "Flattening transactions for status {0}...".format(
                    transaction_status
                )
            )  # noqa
            logger.info(
                "Number of transactions with status {0}: {1}".format(
                    transaction_status, len(transactions_info)
                )
            )
            for transaction in transactions_info:
                transaction["transactionStatus"] = transaction_status
                transaction["transactionCurrency"] = transaction[
                    "transactionAmount"
                ]["currency"]
                transaction["transactionAmount"] = transaction[
                    "transactionAmount"
                ]["amount"]
                if transaction.get("debtorAccount"):
                    transaction["debtorAccountIban"] = transaction[
                        "debtorAccount"
                    ]["iban"]
                    del transaction["debtorAccount"]
                else:
                    transaction["creditorAccountIban"] = transaction[
                        "creditorAccount"
                    ]["iban"]
                    del transaction["creditorAccount"]
                transaction["balanceAfterTransactionAmount"] = transaction[
                    "balanceAfterTransaction"
                ]["balanceAmount"]["amount"]
                transaction["balanceAfterTransactionCurrency"] = transaction[
                    "balanceAfterTransaction"
                ]["balanceAmount"]["currency"]
                del transaction["remittanceInformationUnstructured"]
                del transaction["balanceAfterTransaction"]
                transactions_flatten.append(
                    {self.camel_to_snake(k): v for k, v in transaction.items()}
                )
        logger.info("Finished flattening transactions!")
        return transactions_flatten

    def format_balances(self, balances):
        balances_flattened = balances.get("balances", {})
        balances_flattened = [
            {
                "balanceAmount": balance.get("balanceAmount", {}).get(
                    "amount", ""
                ),
                "balanceCurrency": balance.get("balanceAmount", {}).get(
                    "currency", ""
                ),
                "balanceType": balance.get("balanceType", ""),
            }
            for balance in balances_flattened
        ]
        balances_flattened = [
            {self.camel_to_snake(key): value for key, value in balance.items()}
            for balance in balances_flattened
        ]
        return balances_flattened

    def format_details(self, details):
        details_flattened = details.get("account", {})
        details_flattened = {
            self.camel_to_snake(k): v for k, v in details_flattened.items()
        }
        return details_flattened

    def format_metadata(self, metadata):
        return metadata



import sys
import os
from loguru import logger
from datetime import datetime
from typing import Dict, List, Union

here = os.path.dirname(__file__)

sys.path.append(os.path.join(here, ".."))

from .data_loader import DataLoader  # noqa: E402
from ..data_connections import connect  # noqa: E402


class ExchangeRatesDataLoader(DataLoader):
    NAME: str = "exchange_rates"

    def __init__(self):
        """
        Constructor for ExchangeRatesDataLoader class.
        """
        self.currency_rates_api_connection = connect(self.NAME).connection

    def read(
        self, partition_id: str = None, base_currency: str = "PLN"
    ) -> List[Dict[str, Union[str, float]]]:
        """
        Method for reading data from the exchange rates API.
        """
        logger.info("Reading data from exchange rates API...")
        output = {}
        if partition_id:
            partition_id = datetime(
                int(partition_id[:4]),
                int(partition_id[4:6]),
                int(partition_id[6:]),
            )
        currencies = self.currency_rates_api_connection.get_rates(
            base_currency, date_obj=partition_id
        )
        output = []
        for currency, rate in currencies.items():
            currency_map = {
                "currency": currency,
                "rate": rate,
            }
            output.append(currency_map)
        output.append(
            {
                "currency": base_currency,
                "rate": 1,
            }
        )
        logger.info("Finished reading {0} exchange rates!".format(len(output)))
        return output

    def write(self):
        """
        Method for writing data to the exchange rates API.
        """
        raise NotImplementedError(
            "Exchange rates data loader doesn't support writing data."
        )



from collections import UserDict
import yaml
import os
from pathlib import Path
from typing import Dict, List
import pyspark.sql.types as T


def build_layer_path(layer_name: str) -> str:
    """
    Method for building the path to the layer.

    :param layer_name: The name of the layer.
    :return: The path to the layer.
    """
    return (
        Path(__file__)
        .parent.joinpath("layers")
        .joinpath(layer_name)
        .absolute()
        .as_posix()
    )


# noqa


class Datalake(UserDict):
    INGEST_LAYER_PATH = build_layer_path("ingest")
    BRONZE_LAYER_PATH = build_layer_path("bronze")
    SILVER_LAYER_PATH = build_layer_path("silver")
    GOLD_LAYER_PATH = build_layer_path("gold")
    MASTER_LAYER_PATH = build_layer_path("master")

    def __read_yaml_to_dict__(self, yaml_path):
        with open(yaml_path, "r") as yaml_file:
            yaml_dict = yaml.safe_load(yaml_file)
        return yaml_dict

    def __dict_to_spark_schema__(self, schema: List[Dict[str, str]]):
        """
        Converts a dictionary to a spark schema.

        :param schema: The schema to convert.
        :return: The spark schema.
        """
        spark_schema = T.StructType()
        for field in schema:
            spark_schema.add(field["name"], field["type"])
        return spark_schema

    def __parse_spark_options__(self, options: List[Dict[str, str]]):
        """
        Parses the spark options.

        :param options: The options to parse.
        :return: The parsed options.
        """
        result = {}
        for option in options:
            result[option["name"]] = option["value"]
        return result

    def __setitem__(self, key, item):
        self.__dict__[key] = item

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key]

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        return self.__dict__.copy()

    def has_key(self, k):
        return k in self.__dict__

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def pop(self, *args):
        return self.__dict__.pop(*args)

    def __cmp__(self, dict_):
        return self.__cmp__(self.__dict__, dict_)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __init__(self):
        """
        Constructor for Datalake class.
        """
        result = {}
        for layer_name, layer_path in zip(
            ["ingest", "bronze", "silver", "gold", "master"],
            [
                self.INGEST_LAYER_PATH,
                self.BRONZE_LAYER_PATH,
                self.SILVER_LAYER_PATH,
                self.GOLD_LAYER_PATH,
                self.MASTER_LAYER_PATH,
            ],
        ):
            result[layer_name] = {}
            for file in os.listdir(layer_path):
                if file.endswith(".yaml") or file.endswith(".yml"):
                    loaded_file = self.__read_yaml_to_dict__(
                        os.path.join(layer_path, file)
                    )
                    loaded_file["datalake_layer"] = layer_name
                    result[layer_name][loaded_file["datalake_key"]] = {
                        "datalake_key": loaded_file["datalake_key"],
                        "datalake_bucket": loaded_file["datalake_bucket"],
                        "file_extension": loaded_file["file_extension"],
                    }
                    if loaded_file.get("schema"):
                        result[layer_name][loaded_file["datalake_key"]][
                            "spark_schema"
                        ] = self.__dict_to_spark_schema__(
                            loaded_file["schema"]
                        )
                    if loaded_file.get("options"):
                        result[layer_name][loaded_file["datalake_key"]][
                            "options"
                        ] = self.__parse_spark_options__(
                            loaded_file["options"]
                        )

        self.update(result)



import os
from aws_cdk import App
from aws_cdk import Environment
from cdk_infrastructure.lambdas.raw_to_bronze_stack import (
    RawToBronzeLambdaStack,
)
from cdk_infrastructure.lambdas.constants import RAW_TO_BRONZE_TABLE_NAMES
from cdk_infrastructure.lambdas.ingestion_stack import IngestionLambdaStack
from cdk_infrastructure.lambdas.master_exchange_rates_stack import (
    ExchangeRatesLambdaStack,
)
from cdk_infrastructure.s3_buckets.stack import S3DeployStack
from cdk_infrastructure.glue.stack import (
    GlueDataCatalogStack,
    GlueCrawlersStack,
)
from cdk_infrastructure.s3_buckets.constants import BUCKET_NAMES
from cdk_infrastructure.ecr.budget_guard_ecr_stack import (
    BudgetGuardECRStack,
    BudgetGuardEMRECRStack,
)
from dotenv import load_dotenv

load_dotenv()

env = Environment(
    account=os.environ.get("AWS_ACCOUNT_ID"),
    region=os.environ.get("AWS_REGION_NAME"),
)

app = App()

# Create Lambdas
IngestionLambdaStack(
    app, "LambdaIngestionStack", image_name="budget-guard", env=env
)

for table_name in RAW_TO_BRONZE_TABLE_NAMES:
    RawToBronzeLambdaStack(
        app,
        f"RawToBronzeLambda{table_name.title()}Stack",
        table_name=table_name,
        image_name="budget-guard",
        env=env,
    )

ExchangeRatesLambdaStack(
    app,
    "ExchangeRatesLambdaStack",
    image_name="budget-guard",
    env=env,
)

# Create buckets
for bucket_name in BUCKET_NAMES:
    S3DeployStack(app, f"{bucket_name}Stack", bucket_id=bucket_name, env=env)

S3DeployStack(
    app,
    "BudgetGuardMainCodeStack",
    bucket_id="budget-guard-main",
    env=env,
)

S3DeployStack(
    app,
    "BudgetGuardMasterCodeStack",
    bucket_id="budget-guard-master",
    env=env,
)

# Create AWS Glue Data Catalog
GlueDataCatalogStack(app, "GlueDataCatalogStack", env=env)

# Create AWS Glue Crawlers
GlueCrawlersStack(
    app,
    "GlueCrawlersStack",
    env=env,
)

# Create ECR
BudgetGuardECRStack(app, "BudgetGuardECRStack", env=env)
BudgetGuardEMRECRStack(app, "BudgetGuardEMRECRStack", env=env)

app.synth()



import sys
import os
from loguru import logger
from typing import List, Dict

here = os.path.dirname(__file__)

sys.path.append(os.path.join(here, ".."))

from .raw_to_bronze_pipeline import RawToBronzePipeline  # noqa: E402


class RawToBronzeBalancesPipeline(RawToBronzePipeline):
    INPUT_KEY = "balances"
    OUTPUT_KEY = "balances"

    def transform(self, source_data: List[Dict[str, str]]):
        """
        Transforms the data.

        :param source_data: The data to transform.
        :return: The transformed data.
        """
        logger.info("Transforming data.")
        return [self.format_balances(data) for data in source_data]

    def format_balances(self, balances):
        """
        Formats the balances.

        :param balances: The balances to format.
        :return: The formatted balances.
        """
        balances_flattened = balances.get("balances", {})
        balances_flattened = [
            {
                "balanceAmount": balance.get("balanceAmount", {}).get(
                    "amount", ""
                ),
                "balanceCurrency": balance.get("balanceAmount", {}).get(
                    "currency", ""
                ),
                "balanceType": balance.get("balanceType", ""),
            }
            for balance in balances_flattened
        ]
        balances_flattened = [
            {self.camel_to_snake(key): value for key, value in balance.items()}
            for balance in balances_flattened
        ]
        return balances_flattened



import sys
import os
from typing import Dict, List, Union
from datetime import datetime
from forex_python.converter import CurrencyRates
from loguru import logger
from pyspark.sql import DataFrame as SparkDataFrame
from abc import abstractmethod

here = os.path.dirname(__file__)

sys.path.append(os.path.join(here, ".."))

from ..pipeline import Pipeline  # noqa: E402
from ...datalake import Datalake  # noqa: E402
from ...data_access_service.data_loaders import (  # noqa: E402
    create_data_loader,
)


class BronzeToSilverPipeline(Pipeline):
    INPUT_DATA_LOADER = "spark_s3"
    OUTPUT_DATA_LOADER = "spark_s3"
    INPUT_LAYER = "bronze"
    OUTPUT_LAYER = "silver"

    def __init__(self, partition_id: str) -> None:
        self.datalake = Datalake()
        self.partition_id = partition_id
        self.input_loader = create_data_loader(self.INPUT_DATA_LOADER)
        self.output_loader = create_data_loader(self.OUTPUT_DATA_LOADER)

    def read_sources(self) -> SparkDataFrame:
        """
        Reads the data from the data sources.

        :return: The data from the data sources.
        """
        logger.info("Reading data from datalake.")
        source_df = self.input_loader.read(
            self.datalake[self.INPUT_LAYER][self.INPUT_KEY],
            {
                "partition_id": self.partition_id,
            },
        )
        return source_df

    def write_sources(self, transformed_df: SparkDataFrame):
        """
        Writes the data to the data sources.

        :param transformed_df: The transformed data.
        """
        logger.info("Writing data to datalake.")
        self.output_loader.write(
            transformed_df,
            self.datalake[self.OUTPUT_LAYER][self.OUTPUT_KEY],
            {"partition_id": self.partition_id},
        )

    @abstractmethod
    def transform(self, source_df: SparkDataFrame) -> SparkDataFrame:
        """
        Transforms the data.

        :param source_df: The data to transform.
        :return: The transformed data.
        """
        raise NotImplementedError("Transform method not implemented!")

    def run(self):
        """
        Runs the pipeline.
        """
        logger.info("Running the bronze to silver pipeline...")
        source_df = self.read_sources()
        transformed_df = self.transform(source_df)
        self.write_sources(transformed_df)

    def __get_currency_rates__(
        self,
        base_currency: str = "PLN",
        currencies: List[str] = ["EUR", "GBP", "PLN"],
    ) -> Dict[str, Union[str, Dict[str, float]]]:
        """
        Returns currency rates for given currencies.

        :param base_currency: The base currency.
        :param currencies: The list of currencies to get rates for.
        :return: The currency rates for given currencies.
        """
        logger.info(
            f"Getting currency rates from {','.join(currencies)} to {base_currency}."  # noqa: E501
        )
        partition_id_datetime = datetime(
            int(self.partition_id[:4]),
            int(self.partition_id[4:6]),
            int(self.partition_id[6:]),
        )
        currency_rates_object = CurrencyRates()
        currency_rates = {}
        for currency in currencies:
            currency_rates[currency] = currency_rates_object.get_rate(
                currency, base_currency, partition_id_datetime
            )
        return currency_rates



from .data_loader import DataLoader
from ..data_connections import connect
from loguru import logger
from typing import Dict
from pyspark.sql import DataFrame
import pyspark.sql.functions as F
import pyspark.sql.types as T


class SparkS3DataLoader(DataLoader):
    NAME = "spark_s3"

    def __init__(self):
        """
        Constructor for SparkS3DataLoader class.
        """
        super().__init__()
        self.spark_s3_connection = connect(self.NAME)

    def __build_file_path__(
        self, datalake_config: Dict[str, str], partition_config: Dict[str, str]
    ):
        """
        Method for building the file path.
        """
        bucket_prefix = self.__get_bucket_prefix__()
        return "{0}{1}/{2}/{3}".format(
            bucket_prefix,
            datalake_config["datalake_bucket"],
            datalake_config["datalake_key"],
            self.build_partition_path(partition_config),
        )

    def __get_bucket_prefix__(self) -> str:
        """
        Method for getting the bucket prefix.
        """
        if (
            self.spark_s3_connection.platform
            == self.spark_s3_connection.__Platform__.LOCAL
        ):
            return "s3a://"
        elif (
            self.spark_s3_connection.platform
            == self.spark_s3_connection.__Platform__.EMR
        ):
            return "s3://"
        else:
            raise Exception("Unknown platform!")

    def __build_options__(self, raw_options: Dict[str, str]) -> Dict[str, str]:
        """
        Method for building the options.
        """
        options = {}
        for key, value in raw_options.items():
            if key == "basePath":
                options[key] = self.__build_base_path__(value)
            else:
                options[key] = value
        return options

    def __build_base_path__(self, raw_base_path: str) -> str:
        """
        Method for building the base path.
        """
        bucket_prefix = self.__get_bucket_prefix__()
        return "{0}{1}".format(bucket_prefix, raw_base_path)

    def read(
        self, datalake_config: Dict[str, str], partition_config: Dict[str, str]
    ):
        """
        Method for reading data from the datalake.
        """
        file_path = self.__build_file_path__(datalake_config, partition_config)
        logger.info("Reading data from path: {0}".format(file_path))
        options = self.__build_options__(datalake_config.get("options", {}))
        schema = datalake_config.get("spark_schema", None)
        if schema:
            df = (
                self.spark_s3_connection.spark_session.read.format(
                    datalake_config["file_extension"]
                )
                .options(**options)
                .schema(schema)
                .load(file_path)
            )
        else:
            df = (
                self.spark_s3_connection.spark_session.read.format(
                    datalake_config["file_extension"]
                )
                .options(**options)
                .load(file_path)
            )
        logger.info("Finished reading data from path: {0}".format(file_path))
        return df

    def write(
        self,
        dataframe,
        datalake_config: Dict[str, str],
        partition_config: Dict[str, str],
    ):
        """
        Method for writing data to the datalake.
        """
        file_path = self.__build_file_path__(datalake_config, partition_config)
        logger.info("Writing data to path: {0}".format(file_path))
        options = self.__build_options__(datalake_config.get("options", {}))
        dataframe = self.__apply_schema_on_write__(
            dataframe, datalake_config.get("spark_schema", None)
        )
        (
            dataframe.write.format(datalake_config["file_extension"])
            .options(**options)
            .mode("overwrite")
            .save(file_path)
        )

    def __apply_schema_on_write__(
        self, df: DataFrame, schema: T.StructType
    ) -> DataFrame:
        """
        Method for applying schema on write.
        """
        return df.select(
            [
                F.col(field.name).cast(field.dataType)
                for field in schema.fields
                if field.name in df.columns
            ]
        )



import sys
import os
from loguru import logger
from pyspark.sql import DataFrame as SparkDataFrame

here = os.path.dirname(__file__)

sys.path.append(os.path.join(here, ".."))

from .bronze_to_silver_pipeline import BronzeToSilverPipeline  # noqa: E402


class BronzeToSilverDetailsPipeline(BronzeToSilverPipeline):
    INPUT_DATA_LOADER = "spark_s3"
    OUTPUT_DATA_LOADER = "spark_s3"
    INPUT_LAYER = "bronze"
    OUTPUT_LAYER = "silver"
    INPUT_KEY = "details"
    OUTPUT_KEY = "details"

    def transform(self, source_df: SparkDataFrame) -> SparkDataFrame:
        """
        Transforms the data.

        :param source_df: The data to transform.
        :return: The transformed data.
        """
        logger.info("Transforming data.")
        transformed_df = source_df
        return transformed_df



from pyspark.sql import SparkSession
from pyspark.sql import Row

# Create a Spark session
spark = SparkSession.builder.appName("PySparkTest").getOrCreate()

# Define data as a list of lists
data = [
    [1, "Alice", 28, 50000],
    [2, "Bob", 35, 60000],
    [3, "Charlie", 22, 45000],
    [4, "David", 29, 55000],
    [5, "Eve", 31, 70000],
]

# Convert the list of lists to a list of Row objects
rows = [Row(id=row[0], name=row[1], age=row[2], salary=row[3]) for row in data]

# Create a DataFrame from the list of Row objects
df = spark.createDataFrame(rows)

# Show the schema of the DataFrame
df.printSchema()

# Show the content of the DataFrame
df.show()

# Perform some basic operations
num_rows = df.count()
avg_salary = df.selectExpr("avg(salary)").collect()[0][0]

print(f"Number of rows: {num_rows}")
print(f"Average salary: {avg_salary}")

# Stop the Spark session
spark.stop()



from .aws_connection import AWSConnection
from .connection import Connection
from nordigen import NordigenClient
from nordigen.api import AccountApi
import json
import os
from typing import List, Dict
from loguru import logger


class NordigenConnection(Connection):
    def __init__(self) -> None:
        super().__init__()
        self.aws_connection: AWSConnection = AWSConnection()
        self.nordigen_json_credentials: Dict[str, str] = json.loads(
            self.aws_connection.get_aws_secret("budget_guard_nordigen_key")
        )
        self.country_code: str = os.environ.get("NORDIGEN_COUNTRY_CODE")
        self.accounts: List[AccountApi] = self.connect()

    def connect(self) -> List[AccountApi]:
        logger.info("Connecting to Nordigen...")
        client = NordigenClient(
            secret_id=self.nordigen_json_credentials["secret_id"],
            secret_key=self.nordigen_json_credentials["secret_key"],
            timeout=99999,
        )
        _ = client.generate_token()  # noqa
        logger.info("Retrieving accounts from Nordigen...")
        accounts = client.requisition.get_requisition_by_id(
            requisition_id=self.nordigen_json_credentials["requisition_id"]
        )
        account_ids = accounts["accounts"]
        account_objects = [
            client.account_api(id=account_id) for account_id in account_ids
        ]
        return account_objects



from .aws_connection import AWSConnection
import os
import boto3


class S3Connection(AWSConnection):
    def __init__(self) -> None:
        super().__init__()
        self.s3_client: boto3.client = self.connect()

    def connect(self) -> boto3.client:
        """
        Method to retrieve an S3 client.

        :return: The S3 client.
        """
        client: boto3.client = self.session.client(
            service_name="s3",
            region_name=os.environ.get("AWS_REGION_NAME"),
        )
        return client



from .data_loader import DataLoader
from ..data_connections import connect, NordigenConnection
from loguru import logger
from datetime import datetime
from tqdm import tqdm


class NordigenDataLoader(DataLoader):
    NAME = "nordigen"

    def __init__(self):
        """
        Constructor for NordigenDataLoader class.
        """
        self.nordigen_connection: NordigenConnection = connect(self.NAME)
        self.accounts = self.nordigen_connection.accounts

    def read(self, partition_id: str = None):
        """
        Method for reading data from the Nordigen API.
        """
        logger.info("Reading data from Nordigen...")
        output = {}
        partition_id = self._format_partition_id_for_transactions(partition_id)
        pbar = tqdm(self.accounts, desc="Reading accounts data...")
        for account in pbar:
            meta_data = account.get_metadata()
            pbar.set_description(
                "Reading data for account {0}...".format(meta_data["id"])
            )
            details = account.get_details()
            balances = account.get_balances()
            transactions = account.get_transactions(
                date_from=partition_id, date_to=partition_id
            )
            output[meta_data["id"]] = {
                "details": details,
                "balances": balances,
                "transactions": transactions,
                "metadata": meta_data,
            }
        logger.info("Finished reading {0} accounts data!".format(len(output)))
        return output

    def write(self):
        """
        Method for writing data to the Nordigen API.
        """
        raise NotImplementedError(
            "Nordigen data loader doesn't support writing data."
        )

    def _format_partition_id_for_transactions(
        self, partition_id: str = None
    ) -> str:
        """
        Method for formatting the partition id for transactions.
        """
        if partition_id:
            return datetime.strptime(partition_id, "%Y%m%d").strftime(
                "%Y-%m-%d"
            )
        else:
            return partition_id



from aws_cdk import Stack, RemovalPolicy
from constructs import Construct
from aws_cdk import aws_ecr as ecr


class BudgetGuardECRStack(Stack):
    def __init__(self, scope: Construct, id: str, **kwargs) -> None:
        super().__init__(scope, id, **kwargs)

        # Create an ECR repository
        self.repository = ecr.Repository(
            self,
            "BudgetGuardRepository",
            repository_name="budget-guard",
            removal_policy=RemovalPolicy.DESTROY,
            auto_delete_images=True,
        )

        self.repository.add_lifecycle_rule(
            max_image_count=20,
            description="Remove old images",
            tag_prefix_list=["dev-"],
            rule_priority=1,
        )


class BudgetGuardEMRECRStack(Stack):
    def __init__(self, scope: Construct, id: str, **kwargs) -> None:
        super().__init__(scope, id, **kwargs)

        # Create an ECR repository
        self.repository = ecr.Repository(
            self,
            "BudgetGuardEMRRepository",
            repository_name="budget-guard-emr",
            removal_policy=RemovalPolicy.DESTROY,
            auto_delete_images=True,
        )

        self.repository.add_lifecycle_rule(
            max_image_count=20,
            description="Remove old images",
            tag_prefix_list=["dev-"],
            rule_priority=1,
        )



from .aws_connection import AWSConnection  # noqa
from .connection import Connection  # noqa
from .nordigen_connection import NordigenConnection  # noqa
from .s3_connection import S3Connection  # noqa
from .spark_s3_connection import SparkS3Connection  # noqa
from .exchange_rates_api_connection import ExchangeRatesAPIConnection  # noqa


def connect(connection_name: str) -> Connection:
    """
    Factory function for creating connections.

    :param connection_name: The name of the connection to create.

    :return: The connection.
    """
    if connection_name == "aws":
        return AWSConnection()
    elif connection_name == "s3":
        return S3Connection()
    elif connection_name == "nordigen":
        return NordigenConnection()
    elif connection_name == "spark_s3":
        return SparkS3Connection()
    elif connection_name == "exchange_rates":
        return ExchangeRatesAPIConnection()
    else:
        raise ValueError(f"Connection type {connection_name} not supported.")



import sys
import os
from typing import Dict, List, Union
from loguru import logger
import json

here = os.path.dirname(__file__)

sys.path.append(os.path.join(here, ".."))

from ..pipeline import Pipeline  # noqa: E402
from ...datalake import Datalake  # noqa: E402
from ...data_access_service.data_loaders import (  # noqa: E402
    create_data_loader,
)


class ExchangeRatesPipeline(Pipeline):
    INPUT_DATA_LOADER = "exchange_rates"
    OUTPUT_DATA_LOADER = "s3"
    OUTPUT_LAYER = "master"
    OUTPUT_KEY = "exchange_rates"

    def __init__(self, partition_id: str) -> None:
        self.datalake = Datalake()
        self.partition_id = partition_id
        self.input_loader = create_data_loader(self.INPUT_DATA_LOADER)
        self.output_loader = create_data_loader(self.OUTPUT_DATA_LOADER)

    def read_sources(self) -> List[Dict[str, Union[str, float]]]:
        """
        Reads the data from the data sources.

        :return: The data from the data sources.
        """
        logger.info("Reading data from datalake.")
        source_data = self.input_loader.read(self.partition_id)
        return source_data

    def write_sources(
        self, transformed_data: List[Dict[str, Union[str, float]]]
    ):
        """
        Writes the data to the data sources.

        :param transformed_df: The transformed data.
        """
        logger.info("Writing data to datalake.")
        self.output_loader.write(
            json.dumps(transformed_data),
            self.datalake[self.OUTPUT_LAYER][self.OUTPUT_KEY],
            {"partition_id": self.partition_id},
        )

    def transform(
        self, data: List[Dict[str, Union[str, float]]]
    ) -> List[Dict[str, Union[str, float]]]:
        """
        Transforms the data.

        :param data: The data to transform.
        :return: The transformed data.
        """
        logger.info("Transforming data.")
        return data

    def run(self):
        """
        Runs the pipeline.
        """
        data = self.read_sources()
        transformed_data = self.transform(data)
        self.write_sources(transformed_data)
